CREATE OR REPLACE VIEW fw_sys_stat_schema (version) AS
	SELECT '4.0.0'
	FROM dual;

--Enable reinstallation of FW Components:
CREATE TABLE FW_SYS_STAT_FORCE_REINSTALL
   		     (ID NUMBER,
   		      PROJECT_ID NUMBER NOT NULL,
   		      COMPONENT_ID NUMBER NOT NULL,
   		      DESC_FILE VARCHAR2(1024) NOT NULL,
   		      RESTART_REQUIRED NUMBER DEFAULT 0,
   		      OVERWRITE_FILES NUMBER DEFAULT 0,
   		      EXECUTED_ON DATE,
   		      CONSTRAINT REINSTALL_ID_PK PRIMARY KEY (ID));

ALTER TABLE FW_SYS_STAT_FORCE_REINSTALL
	ADD CONSTRAINT PROJECT_ID_FK
	FOREIGN KEY (PROJECT_ID)
	REFERENCES FW_SYS_STAT_PVSS_PROJECT(ID) ON DELETE CASCADE;

ALTER TABLE FW_SYS_STAT_FORCE_REINSTALL
	ADD CONSTRAINT COMPONENT_ID_FK
	FOREIGN KEY (COMPONENT_ID)
	REFERENCES FW_SYS_STAT_COMPONENT(ID) ON DELETE CASCADE;

CREATE SEQUENCE FW_SYS_STAT_REINSTALL_SQ
    MINVALUE 1
    MAXVALUE 9999999
    START WITH 1
    INCREMENT BY 1
    NOCACHE;

ALTER TABLE FW_SYS_STAT_PVSS_SYSTEM
ADD COMPUTER_ID NUMBER;

ALTER TABLE FW_SYS_STAT_PVSS_PROJECT
ADD REDU_COMPUTER_ID NUMBER;

ALTER TABLE FW_SYS_STAT_PVSS_PROJECT
	ADD CONSTRAINT REDU_ID_FK
	FOREIGN KEY (REDU_COMPUTER_ID)
	REFERENCES FW_SYS_STAT_COMPUTER(ID) ON DELETE CASCADE;

ALTER TABLE fw_sys_stat_proj_install_comps
ADD COMPUTER_ID NUMBER;

ALTER TABLE fw_sys_stat_proj_install_comps
	ADD CONSTRAINT INST_COMP_PC_ID_FK
	FOREIGN KEY (COMPUTER_ID)
	REFERENCES FW_SYS_STAT_COMPUTER(ID) ON DELETE CASCADE;


--Update project install component tables:
BEGIN
	FOR REC IN (SELECT ID, PROJECT_ID
  			    FROM fw_sys_stat_proj_install_comps
  			    WHERE VALID_UNTIL IS NULL AND COMPUTER_ID IS NULL) LOOP

 		EXECUTE IMMEDIATE ('UPDATE fw_sys_stat_proj_install_comps
 							SET COMPUTER_ID =(SELECT COMPUTER_ID FROM FW_SYS_STAT_PVSS_PROJECT WHERE ID = '|| REC.PROJECT_ID ||')
 							WHERE ID = '||REC.ID);
	END LOOP;
END;
/
commit;

--
--Adding support for Redundant Systems
CREATE OR REPLACE VIEW fw_sys_stat_proj_comps AS
	SELECT HOSTNAME, PROJECT_NAME, COMPONENT_NAME, MAX(COMPONENT_VERSION) COMPONENT_VERSION, IS_SUBCOMPONENT, DEFAULT_PATH, IS_OFFICIAL, IS_PATCH, DESCRIPTION_FILE, OVERWRITE_FILES, FORCE_REQUIRED, IS_SILENT, MAX(VALID_FROM) as M_VALID_FROM, RESTART_PROJECT
	FROM (SELECT PC.HOSTNAME, P.PROJECT_NAME, C.COMPONENT_NAME, C.COMPONENT_VERSION, C.IS_SUBCOMPONENT, C.DEFAULT_PATH, C.IS_OFFICIAL, C.IS_PATCH, GC.DESCRIPTION_FILE, PG.OVERWRITE_FILES, PG.FORCE_REQUIRED, PG.IS_SILENT, GC.VALID_FROM, PG.RESTART_PROJECT
			FROM FW_SYS_STAT_PVSS_PROJECT P, FW_SYS_STAT_GROUP_OF_COMP G, FW_SYS_STAT_PROJECT_GROUPS PG, FW_SYS_STAT_COMP_IN_GROUPS GC, FW_SYS_STAT_COMPONENT C, FW_SYS_STAT_COMPUTER PC
			WHERE P.ID = PG.PROJECT_ID AND PG.VALID_UNTIL IS NULL AND P.VALID_UNTIL IS NULL AND G.ID = PG.GROUP_ID AND GC.GROUP_ID = G.ID AND GC.VALID_UNTIL IS NULL AND C.ID = GC.FW_COMPONENT_ID AND C.VALID_UNTIL IS NULL AND (PC.ID = P.COMPUTER_ID OR PC.ID = P.REDU_COMPUTER_ID) AND PC.VALID_UNTIL IS NULL)
	GROUP BY COMPONENT_NAME, HOSTNAME, PROJECT_NAME, COMPONENT_NAME, IS_SUBCOMPONENT, DEFAULT_PATH, IS_OFFICIAL, IS_PATCH, DESCRIPTION_FILE, OVERWRITE_FILES, FORCE_REQUIRED, IS_SILENT, RESTART_PROJECT
	ORDER BY M_VALID_FROM;

